﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace Sand
{
    public class SandTexture
    {
        Texture2D Result;
        Texture2D ResultingVelocities;

        Texture2D density1;
        Texture2D density2;
        Texture2D velocity1;
        Texture2D velocity2;
        Texture2D palette;
        Texture2D finger;
        Texture2D backdrop;

        Effect advect;
        Effect move;
        Effect display;

        RenderTarget2D vel1;
        RenderTarget2D vel2;
        RenderTarget2D den1;
        RenderTarget2D den2;

        int Width;
        int Height;
        int Pass;

        Rectangle fullscreen;

        public SandTexture(int width,int height)
        {
            Width = width;
            Height = height;

            vel1 = new RenderTarget2D(Game1.Instance.GraphicsDevice, width, height, 1, SurfaceFormat.Vector2);
            vel2 = new RenderTarget2D(Game1.Instance.GraphicsDevice, width, height, 1, SurfaceFormat.Vector2);
            den1 = new RenderTarget2D(Game1.Instance.GraphicsDevice, width, height, 1, SurfaceFormat.Single);
            den2 = new RenderTarget2D(Game1.Instance.GraphicsDevice, width, height, 1, SurfaceFormat.Single);
            Pass = 0;

            Vector2[] initialVelocities = new Vector2[width * height];
            float[] initialDensities = new float[width * height];
            for (int i = 0; i < width * height; i++)
            {
                initialVelocities[i] = Vector2.Zero;
                initialDensities[i] = 1.0f;
            }
            velocity1 = new Texture2D(Game1.Instance.GraphicsDevice, width, height,1,TextureUsage.None,SurfaceFormat.Vector2);
            velocity1.SetData<Vector2>(initialVelocities);
            density1 = new Texture2D(Game1.Instance.GraphicsDevice, width, height,1,TextureUsage.None,SurfaceFormat.Single);
            density1.SetData<float>(initialDensities);

            palette = Game1.Instance.Content.Load<Texture2D>("palette");
            backdrop = Game1.Instance.Content.Load<Texture2D>("Backdrop");

            advect = Game1.Instance.Content.Load<Effect>("Advect");
            move = Game1.Instance.Content.Load<Effect>("Density");
            display = Game1.Instance.Content.Load<Effect>("Display");

            fullscreen = new Rectangle(0, 0, Game1.Instance.GraphicsDevice.Viewport.Width, Game1.Instance.GraphicsDevice.Viewport.Height);


            // generate a texure to use when a mouse click occurs.
            finger = new Texture2D(Game1.Instance.GraphicsDevice, 5, 5, 1, TextureUsage.None, SurfaceFormat.Vector2);
            Vector2[] fingerdata = new Vector2[5 * 5];
            int index = 0;
            for (int y = -2; y < 3; y++)
            {
                for (int x = -2; x < 3; x++)
                {
                    Vector2 vel = new Vector2();
                    if ((x != 0) && (y != 0))
                    {
                        vel.X = x;
                        vel.Y = y;
                        vel /= vel.LengthSquared();
                    }
                    else
                    {
                        vel.X = -1;
                        vel.Y = -1;
                    }
                    fingerdata[index++] = vel;
                }
            }
            finger.SetData<Vector2>(fingerdata);
            
        }

        public void Update(GameTime gt)
        {
            float dt = gt.ElapsedGameTime.Milliseconds / 1000.0f;

            MouseState ms = Mouse.GetState();
            Vector2 mousepos = new Vector2();
            mousepos.X = (ms.X * Width) / Game1.Instance.GraphicsDevice.Viewport.Width;
            mousepos.Y = (ms.Y * Height) / Game1.Instance.GraphicsDevice.Viewport.Height;
            advect.Parameters["timestep"].SetValue(dt);
            advect.Parameters["rdx"].SetValue(1.0f / (float)Width);
            advect.Parameters["dissipation"].SetValue(0.95f);
            move.Parameters["timestep"].SetValue(dt);
            move.Parameters["rdx"].SetValue(1.0f / Width);
            move.Parameters["dissipation"].SetValue(1.0f);
            switch (Pass)
            {
                case 0:
                    Game1.Instance.GraphicsDevice.SetRenderTarget(0, vel2);
                    Game1.Instance.spriteBatch.Begin(SpriteBlendMode.None, SpriteSortMode.Immediate, SaveStateMode.None);
                    advect.Parameters["Map"].SetValue(velocity1);
                    advect.CommitChanges();
                    advect.Begin();
                    advect.CurrentTechnique.Passes[0].Begin();
                    Game1.Instance.spriteBatch.Draw(velocity1, Vector2.Zero, Color.White);
                    Game1.Instance.spriteBatch.End();
                    advect.CurrentTechnique.Passes[0].End();
                    advect.End();
                    if (ms.LeftButton == ButtonState.Pressed)
                    {
                        Game1.Instance.spriteBatch.Begin();
                        Game1.Instance.spriteBatch.Draw(finger, mousepos, Color.White);
                        Game1.Instance.spriteBatch.End();
                        
                    }
                    Game1.Instance.GraphicsDevice.SetRenderTarget(0, den2);
                    velocity2 = vel2.GetTexture();
                    ResultingVelocities = velocity2;

                    Game1.Instance.spriteBatch.Begin(SpriteBlendMode.None, SpriteSortMode.Immediate, SaveStateMode.None);
                    move.Parameters["VelocityMap"].SetValue(velocity2);
                    move.Parameters["DensityMap"].SetValue(density1);
                    move.CommitChanges();
                    move.Begin();
                    move.CurrentTechnique.Passes[0].Begin();
                    Game1.Instance.spriteBatch.Draw(density1, Vector2.Zero, Color.White);
                    Game1.Instance.spriteBatch.End();
                    move.CurrentTechnique.Passes[0].End();
                    move.End();
                    Game1.Instance.GraphicsDevice.SetRenderTarget(0, null);
                    density2 = den2.GetTexture();
                    Result = density2;
                    Pass = 1;
                    break;

                case 1:
                    Game1.Instance.GraphicsDevice.SetRenderTarget(0, vel1);
                    advect.Parameters["Map"].SetValue(velocity2);
                    advect.CommitChanges();
                    Game1.Instance.spriteBatch.Begin(SpriteBlendMode.None, SpriteSortMode.Immediate, SaveStateMode.None);
                    advect.Begin();
                    advect.CurrentTechnique.Passes[0].Begin();
                    Game1.Instance.spriteBatch.Draw(velocity2, Vector2.Zero, Color.White);
                    Game1.Instance.spriteBatch.End();
                    advect.CurrentTechnique.Passes[0].End();
                    advect.End();
                    if (ms.LeftButton == ButtonState.Pressed)
                    {
                        Game1.Instance.spriteBatch.Begin();
                        Game1.Instance.spriteBatch.Draw(finger, mousepos, Color.White);
                        Game1.Instance.spriteBatch.End();

                    } 
                    Game1.Instance.GraphicsDevice.SetRenderTarget(0, den1);
                    velocity1 = vel1.GetTexture();
                    ResultingVelocities = velocity1;

                    Game1.Instance.spriteBatch.Begin(SpriteBlendMode.None, SpriteSortMode.Immediate, SaveStateMode.None);
                    move.Parameters["VelocityMap"].SetValue(velocity1);
                    move.Parameters["DensityMap"].SetValue(density2);
                    move.CommitChanges();
                    move.Begin();
                    move.CurrentTechnique.Passes[0].Begin();
                    Game1.Instance.spriteBatch.Draw(density2, Vector2.Zero, Color.White);
                    Game1.Instance.spriteBatch.End();
                    move.CurrentTechnique.Passes[0].End();
                    move.End();
                    Game1.Instance.GraphicsDevice.SetRenderTarget(0, null);
                    density1 = den1.GetTexture();
                    Result = density1;
                    Pass = 0;
                    break;
            }
        }

        public void Draw()
        {
            Game1.Instance.spriteBatch.Begin();
            Game1.Instance.spriteBatch.Draw(backdrop, fullscreen, Color.White);
            Game1.Instance.spriteBatch.End();
            
            Game1.Instance.spriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Immediate, SaveStateMode.None);
            display.Parameters["DensityMap"].SetValue(Result);
            display.Parameters["PaletteMap"].SetValue(palette);
            display.Begin();
            display.CurrentTechnique.Passes[0].Begin();
            Game1.Instance.spriteBatch.Draw(Result, fullscreen, Color.White);
            Game1.Instance.spriteBatch.End();
            display.CurrentTechnique.Passes[0].End();
            display.End();

#if DEBUG
            Game1.Instance.spriteBatch.Begin();
            Game1.Instance.spriteBatch.Draw(ResultingVelocities, Vector2.Zero, Color.White);
            Game1.Instance.spriteBatch.Draw(Result, new Vector2(Game1.Instance.GraphicsDevice.Viewport.Width - Width, 0), Color.White);
            Game1.Instance.spriteBatch.End();
         
#endif

        }
    }
}
